@@ -21,39 +21,17 @@
 (defun some-fixnum-p (temp x y)
   (inst mov :dword temp x)
   (inst and :dword temp y)
-  (inst test :byte temp fixnum-tag-mask))
-
-(defun static-fun-addr (name)
-  #+immobile-code (make-fixup name :static-call)
-  #-immobile-code (ea (+ nil-value (static-fun-offset name))))
-
-(defun call-static-fun (fun arg-count)
-  (inst push rbp-tn)
-  (inst mov rbp-tn rsp-tn)
-  (inst sub rsp-tn (* n-word-bytes 2))
-  (inst mov (ea rsp-tn) rbp-tn)
-  (inst mov rbp-tn rsp-tn)
-  (inst mov rcx-tn (fixnumize arg-count))
-  (inst call (static-fun-addr fun))
-  (inst pop rbp-tn))
-
-(defun tail-call-static-fun (fun arg-count)
-  (inst push rbp-tn)
-  (inst mov rbp-tn rsp-tn)
-  (inst sub rsp-tn n-word-bytes)
-  (inst push (ea (frame-byte-offset return-pc-save-offset) rbp-tn))
-  (inst mov rcx-tn (fixnumize arg-count))
-  (inst jmp (static-fun-addr fun))))
-
+  (inst test :byte temp fixnum-tag-mask)))
 
 ;;;; addition, subtraction, and multiplication
 
 #+sb-assembling
 (defun return-single-word-bignum (dest alloc-tn source)
-  (instrument-alloc 16 nil)
-  (let ((header (logior (ash 1 n-widetag-bits) bignum-widetag)))
-    (pseudo-atomic ()
-      (allocation nil 16 0 nil nil alloc-tn)
+  (let ((header (logior (ash 1 n-widetag-bits) bignum-widetag))
+        (nbytes #+bignum-assertions 32 #-bignum-assertions 16))
+    (emit-instrument-alloc nil thread-tn bignum-widetag nbytes alloc-tn)
+    (allocating ()
+      (emit-allocation nil thread-tn bignum-widetag nbytes 0 alloc-tn nil)
       (storew* header alloc-tn 0 0 t)
       (storew source alloc-tn bignum-digits-offset 0)
       (if (eq dest alloc-tn)
@@ -132,7 +110,7 @@
     (inst cmp x rcx)
     (inst jmp :e SINGLE-WORD-BIGNUM)
 
-    (alloc-other res bignum-widetag (+ bignum-digits-offset 2) nil)
+    (emit-alloc-other nil thread-tn bignum-widetag (+ bignum-digits-offset 2) res)
     (storew rax res bignum-digits-offset other-pointer-lowtag)
     (storew rcx res (1+ bignum-digits-offset) other-pointer-lowtag)
     (inst clc) (inst ret)
@@ -149,23 +127,26 @@
                           (:translate %negate)
                           (:save-p t))
                          ((:arg x (descriptor-reg any-reg) rdx-offset)
-                          (:res res (descriptor-reg any-reg) rdx-offset)
-                          (:temp rcx unsigned-reg rcx-offset))
+                          (:res res (descriptor-reg any-reg) rdx-offset))
   (inst test :byte x fixnum-tag-mask)
   (inst jmp :nz GENERIC)
   (move res x)
   (inst neg res)                        ; (- most-negative-fixnum) is BIGNUM
-  (inst jmp :o BIGNUM)
-  (inst clc) (inst ret)
-  BIGNUM
-  (inst shr res n-fixnum-tag-bits)      ; sign bit is data - remove type bits
-  (return-single-word-bignum res rcx res)
+  ;; This constant isn't really a fixup, but it's easiest for me to think about it
+  ;; that way for now. It should really do whatever EMIT-EA does for a CONSTANT.
+  (inst cmov :o res (ea (make-fixup nil :code-object
+                                    (+ (ash code-constants-offset word-shift)
+                                       (- other-pointer-lowtag)))
+                        rip-tn))
   (inst clc) (inst ret)
   GENERIC
   (tail-call-static-fun '%negate 1))
 
 ;;;; comparison
 
+(eval-when (:compile-toplevel)
+  (assert (minusp (static-symbol-offset t))))
+
 (macrolet ((define-cond-assem-rtn (name translate static-fn test)
              `(define-assembly-routine (,name
                                         (:translate ,translate)
@@ -186,12 +167,12 @@
 
                 DO-STATIC-FUN
                 (call-static-fun ',static-fn 2)
-                ;; HACK: We depend on NIL having the lowest address of all
-                ;; static symbols (including T)
+                ;; X now holds T or NIL corresponding to the answer but it needs
+                ;; to be returned as :L or :G in EFLAGS. We rely on address of T
+                ;; being less address of NIL (asserted above)
                 ,@(ecase test
-                    (:l `((inst mov y (1+ nil-value))
-                          (inst cmp y x)))
-                    (:g `((inst cmp x (1+ nil-value))))))))
+                    (:l `((inst cmp x null-tn)))
+                    (:g `((inst cmp null-tn x)))))))
   (define-cond-assem-rtn generic-< < two-arg-< :l)
   (define-cond-assem-rtn generic-> > two-arg-> :g))
 
@@ -214,7 +195,8 @@
 
   DO-STATIC-FUN
   (call-static-fun 'two-arg-= 2)
-  (inst cmp x (+ nil-value (static-symbol-offset t))))
+  (inst sub x null-tn)
+  (inst cmp x (static-symbol-offset t)))
 
 #+sb-assembling
 (define-assembly-routine (logcount)
@@ -300,30 +282,35 @@
                  (inst jmp :ge POSITIVE)
                  (inst not rdx)
                  POSITIVE))
-           (unless (memq :popcnt *backend-subfeatures*)
-             (test-cpu-feature cpu-has-popcnt)
-             (inst jmp :z slow))
+           (when (memq :popcnt *backend-subfeatures*) ; always use POPCNT
            ;; Intel's implementation of POPCNT on some models treats it as
            ;; a 2-operand ALU op in the manner of ADD,SUB,etc which means that
            ;; it falsely appears to need data from the destination register.
            ;; The workaround is to clear the destination.
            ;; See http://stackoverflow.com/questions/25078285
-           (unless (location= result arg)
+             (unless (location= result arg)
              ;; We only break the spurious dep. chain if result isn't the same
              ;; register as arg. (If they're location=, don't trash the arg!)
-             (inst xor result result))
+               (inst xor :dword result result))
+             (inst popcnt result arg)
+             (return-from ,name))
+
+           ;; Conditionally use POPCNT
+           (test-cpu-feature cpu-has-popcnt)
+           (inst jmp :z slow)
+           (unless (location= result arg)
+             (inst xor :dword result result))
            (inst popcnt result arg)
            (inst jmp done)
          slow
-           (unless (memq :popcnt *backend-subfeatures*)
-             (move rdx arg)
-             (invoke-asm-routine 'call 'logcount vop)
-             (move result rdx))
+           (move rdx arg)
+           (invoke-asm-routine 'call 'logcount vop)
+           (move result rdx)
          done))))
   (def-it unsigned-byte-64-count 14 unsigned-reg unsigned-num)
   (def-it signed-byte-64-count 15 signed-reg signed-num :signed t)
   (def-it positive-fixnum-count 12 any-reg positive-fixnum)
-  (def-it positive-fixnum-count 13 any-reg fixnum :signed t))
+  (def-it signed-fixnum-count 13 any-reg fixnum :signed t))
 
 ;;; General case of EQL
 
@@ -345,12 +332,15 @@
 ;;;   (define-assembly-routine ... ((:temp ...)) (:let ((c (register-inline-constant ...)))))
 ;;; It just doesn't seem worth the effort to do all that.
 (defparameter eql-dispatch nil)
-(define-assembly-routine (generic-eql (:return-style :none))
-    ((:temp rax unsigned-reg rax-offset)
-     (:temp rcx unsigned-reg rcx-offset)
+(define-assembly-routine (generic-eql (:return-style :none)
+                                      (:export generic-eql* eql-ratio))
+    ((:temp rcx unsigned-reg rcx-offset)  ; callee-saved
+     (:temp rax unsigned-reg rax-offset)  ; vop temps
      (:temp rsi unsigned-reg rsi-offset)
      (:temp rdi unsigned-reg rdi-offset)
      (:temp r11 unsigned-reg r11-offset))
+  (inst mov :byte rax (ea (- other-pointer-lowtag) rdi))
+  (inst sub :byte rax bignum-widetag)
   ;; SINGLE-FLOAT is included in this table because its widetag within the range
   ;; of accepted widetags in the precondition for calling this routine.
   ;; Technically it would be an error to see an other-pointer object with that tag.
@@ -371,6 +361,7 @@
   ;; widetags are spaced 4 apart, and we've subtracted BIGNUM_WIDETAG,
   ;; so scaling by 2 gets a multiple of 8 with bignum at offset 0 etc.
   ;; But the upper 3 bytes of EAX hold junk presently, so clear them.
+  GENERIC-EQL*
   (inst and :dword rax #x7f)
   (inst lea r11 eql-dispatch)
   (inst jmp (ea r11 rax 2))
@@ -429,21 +420,28 @@
   EQL-BIGNUM
   (inst mov rax (ea (- other-pointer-lowtag) rdi))
   (inst xor rax (ea (- other-pointer-lowtag) rsi))
-  ;; kill the GC mark bit by shifting out (as if we'd actually get fullcgc
-  ;; to concurrently mark the heap. Well, I can hope can't I?)
-  (inst shl rax 1)
   (inst jmp :ne done) ; not equal
-  ;; Preserve rcx and compute the header length in words.
-  ;; Maybe we should deem that bignums have at most (2^32-1) payload words
-  ;; so that I can use an unaligned 'movl'. Or, maybe I should put the length
-  ;; in the high 4 bytes of the header so that the same GC mark bit would work
-  ;; as for all other objects (in byte index 3 of the header)
+  ;; Preserve rcx and compute the length in words.
+  ;; MAXIMUM-BIGNUM-LENGTH is a 4-byte quantity. Probably better if this would use
+  ;; the high 4 bytes of the header to naturally align the load.
   (inst push rcx)
-  (inst mov rcx (ea (- other-pointer-lowtag) rdi))
-  (inst shl rcx 1)
-  (inst shr rcx (1+ n-widetag-bits))
-  (inst jmp :z epilogue) ; zero payload length, can this happen?
+  (inst mov :dword rcx (ea (- 1 other-pointer-lowtag) rdi))
+  #+bignum-assertions
+  (progn (inst mov r11 rcx)
+         (inst or r11 1)) ; align to start of ubsan bits
   compare-loop
+  #+bignum-assertions
+  (let ((ok1 (gen-label)) (ok2 (gen-label)))
+    (inst dec rcx)
+    (inst bt (ea (- n-word-bytes other-pointer-lowtag) rsi r11 8) rcx)
+    (inst jmp :nc ok1)
+    (inst break halt-trap)
+    (emit-label ok1)
+    (inst bt (ea (- n-word-bytes other-pointer-lowtag) rdi r11 8) rcx)
+    (inst jmp :nc ok2)
+    (inst break halt-trap)
+    (emit-label ok2)
+    (inst inc rcx))
   (inst mov rax (ea (- other-pointer-lowtag) rsi rcx 8))
   (inst cmp rax (ea (- other-pointer-lowtag) rdi rcx 8))
   (inst jmp :ne epilogue)
@@ -498,13 +496,14 @@
 ;; the same widetag, but not bignum-widetag, the behavior is undefined.
 ;;
 (define-assembly-routine (%eql/integer
-                          (:translate %eql/integer)
+                          (:translate eql)
                           ;; :safe would imply signaling an error
                           ;; if the args are not integer, which this doesn't.
                           (:policy :fast-safe)
                           (:conditional :e)
-                          (:cost 10))
-                         ((:arg x (descriptor-reg any-reg) rdx-offset)
+                          (:cost 10)
+                          (:arg-types (:or integer bignum) *))
+                         ((:arg x (descriptor-reg) rdx-offset)
                           (:arg y (descriptor-reg any-reg) rdi-offset)
                           (:temp rcx unsigned-reg rcx-offset)
                           (:temp rax unsigned-reg rax-offset))
@@ -522,8 +521,6 @@
   (inst cmp rcx (ea (- other-pointer-lowtag) y))
   (inst jmp :ne done)
   (inst shr rcx n-widetag-bits)
-  ;; can you have 0 payload words? Probably not, but let's be safe here.
-  (inst jmp :z done)
   loop
   (inst mov rax (ea (- other-pointer-lowtag) x rcx 8))
   (inst cmp rax (ea (- other-pointer-lowtag) y rcx 8))
@@ -536,3 +533,9 @@
   (inst jmp :ne loop)
   ;; If the Z flag is set, the integers were EQL
   done)
+
+#-sb-assembling
+(define-vop (%eql/integer2 %eql/integer)
+  (:args (x :scs (descriptor-reg any-reg) :target x-arg-temp)
+         (y :scs (descriptor-reg) :target y-arg-temp))
+  (:arg-types * (:or integer bignum)))
