@@ -951,7 +951,7 @@ between the ~A definition and the ~A definition"
 (declaim (type cons **non-instance-classoid-types**))
 (defglobal **non-instance-classoid-types**
   '(symbol system-area-pointer weak-pointer code-component
-    #-(or x86 x86-64 arm64 riscv) lra
+    #-(or x86 x86-64 arm64 riscv loongarch64) lra
     fdefn random-class))
 
 (defun classoid-non-instance-p (classoid)
@@ -1052,7 +1052,7 @@ between the ~A definition and the ~A definition"
      (code-component :codes (,sb-vm:code-header-widetag)
                      :predicate code-component-p
                      :prototype-form (fun-code-header #'identity))
-     #-(or x86 x86-64 arm64 riscv)
+     #-(or x86 x86-64 arm64 riscv loongarch64)
      (lra :codes (,sb-vm:return-pc-widetag)
           :predicate lra-p
           ;; Make the PROTOTYPE slot unbound.
