@@ -63,6 +63,16 @@
 (defun find-or-create-equality-constraint (operator x y not-p &optional (amount 0))
   (unless amount
     (setf amount 0))
+  (when (and (lambda-var-p x)
+             (lambda-var-p y))
+    ;; Normalize var-var constraints to < and <=
+    (case operator
+      (>
+       (rotatef x y)
+       (setf operator '<))
+      (>=
+       (rotatef x y)
+       (setf operator '<=))))
   (let ((x-var (constraint-var x))
         (cache-key (typecase y
                      (vector-length-constraint y
@@ -894,7 +904,8 @@
 (defoptimizer (make-sequence constraint-propagate-result) ((type length &rest args) node)
   (list (list 'vector-length length)))
 
-(defoptimizers constraint-propagate-result (remove delete remove-if delete-if remove-if-not delete-if-not)
+(defoptimizers constraint-propagate-result (remove delete remove-if delete-if remove-if-not delete-if-not
+                                            copy-remove copy-remove-if-not copy-remove-if)
     ((x sequence &rest args &key from-end test test-not start end count key) node gen)
   (let (c)
     (let ((var (ok-lvar-lambda-var sequence gen)))
@@ -906,6 +917,9 @@
              c)))
     c))
 
+(defoptimizer (random constraint-propagate-result) ((num &optional state) node)
+  (list (list '< num)))
+
 (defun subseq-bounds (sequence start end gen)
   (let* ((null-p (types-equal-or-intersect (lvar-type end) (specifier-type 'null)))
          (end (if null-p
@@ -968,6 +982,19 @@
                         (incf max-sum n)
                         (loop repeat n
                               do (pop args))))
+                     ((and subseq
+                           (constant-lvar-p arg)
+                           (eq (lvar-value arg) 'sb-impl::%repeat))
+                      (let* ((length (pop args))
+                             (int (type-approximate-interval (lvar-type length))))
+                        (cond (int
+                               (incf min-sum (or (interval-low int) 0))
+                               (if (interval-high int)
+                                   (incf max-sum (interval-high int))
+                                   (setf max nil)))
+                              (t
+                               (setf max nil))))
+                      (pop args))
                      (t
                       (let ((int (type-approximate-interval (type-intersection (or (vector-length-type (lvar-type arg))
                                                                                    *universal-type*)
@@ -1280,3 +1307,47 @@
 
 (defoptimizer (array-rank= constraint-propagate-if) ((x n))
   (values x (specifier-type `(array * ,(lvar-value n)))))
+
+(defoptimizer (array-dimensions-equal constraint-propagate-if) ((array1 array2))
+  (let* ((array1-type (lvar-conservative-type array1))
+         (array2-type (lvar-conservative-type array2))
+         (dims1 (array-type-dimensions-or-give-up array1-type nil))
+         (dims2 (array-type-dimensions-or-give-up array2-type nil)))
+    (flet ((try (dims1 dims2)
+             (when (consp dims1)
+               (let* ((new-dims (if (consp dims2)
+                                    (loop for dim1 in dims1
+                                          for dim2 in dims2
+                                          collect (if (eq dim1 '*)
+                                                      dim2
+                                                      dim1))
+                                    dims1))
+                      (new (make-array-type (if (or (conservative-array-type-complexp array1-type)
+                                                    (conservative-array-type-complexp array2-type))
+                                                (make-list (length new-dims) :initial-element '*)
+                                                new-dims)
+                                            :element-type *wild-type*)))
+                 (list (list 'typep array1 new)
+                       (list 'typep array2 new))))))
+      (values nil nil (or (try dims1 dims2)
+                          (try dims2 dims1))))))
+
+(defoptimizer (array-dimensions-equal-list constraint-propagate-if) ((array1 list))
+  (when (constant-lvar-p list)
+    (let ((dims2 (lvar-value list)))
+      (when (and (proper-list-p dims2)
+                 (every #'integerp dims2))
+        (let* ((array1-type (lvar-conservative-type array1))
+               (new (make-array-type (if (conservative-array-type-complexp array1-type)
+                                         (make-list (length dims2) :initial-element '*)
+                                         dims2)
+                                     :element-type *wild-type*)))
+          (values array1 new nil nil t))))))
+
+(defoptimizer (array-dimension constraint-propagate) ((array axis) node gen)
+  (unless (types-equal-or-intersect (lvar-type axis) (specifier-type '(eql 0)))
+    (let ((var (ok-lvar-lambda-var array gen)))
+      (when var
+        (list (list 'typep var
+                    (specifier-type '(and array (not vector)))
+                    nil))))))
