@@ -552,46 +552,42 @@
 (defun find-source-paths (form tlf-num)
   (declare (type index tlf-num))
   (let ((*current-form-number* 0))
-    (sub-find-source-paths form (list tlf-num)))
+    (sub-find-source-paths form (list tlf-num) 0))
   (values))
-(defun sub-find-source-paths (form path)
+(defun sub-find-source-paths (form path depth)
   (unless (get-source-path form)
     (note-source-path form path)
-    (incf *current-form-number*)
+    (unless (< depth 100) ; ARB, see lp#654289
+      #+sb-xc-host (bug "Unexpected depth of code")
+      #-sb-xc-host (return-from sub-find-source-paths))
     (let ((pos 0)
           (subform form)
           (trail form))
       (declare (fixnum pos))
       (macrolet ((frob ()
                    `(progn
-                      (let ((fm (cond ((comma-p subform)
-                                       (comma-expr subform))
-                                      ((atom subform)
-                                       (return))
-                                      (t
-                                       (car subform)))))
-                        (when (comma-p fm)
-                          (setf fm (comma-expr fm)))
-                        (cond ((consp fm)
-                               ;; If it's a cons, recurse.
-                               (sub-find-source-paths fm (cons pos path)))
-                              ((eq 'quote fm)
-                               ;; Don't look into quoted constants.
-                               ;; KLUDGE: this can't actually know about constants.
-                               ;; e.g. (let ((quote (error "foo")))) or
-                               ;; (list quote (error "foo")) are not
-                               ;; constants and yet are ignored.
-                               (return))
-                              ((not (zerop pos))
-                               ;; Otherwise store the containing form. It's not
-                               ;; perfect, but better than nothing.
-                               (note-source-path subform pos path)))
-                        (incf pos))
-                      (when (comma-p subform)
-                        (return))
-                      (setq subform (cdr subform))
+                      (cond
+                        ;; (a b . ,c) -> (a b comma c)
+                        ((comma-p subform)
+                         (setq subform (list 'comma (comma-expr subform))))
+                        ((atom subform) (return)))
+                      (let ((fm (car subform)))
+                        (cond
+                          ((consp fm)
+                           (incf *current-form-number*)
+                           (sub-find-source-paths fm (cons pos path) (1+ depth)))
+                          ;; (a b ,c d) -> (a b (comma c) d)
+                          ((comma-p fm)
+                           (incf *current-form-number*)
+                           (sub-find-source-paths (list 'comma (comma-expr fm)) (cons pos path) (1+ depth)))
+                          ((not (zerop pos))
+                           (unless (get-source-path subform)
+                             (note-source-path subform pos path)))))
+                      (setq subform (cdr subform)
+                            pos (1+ pos))
                       (when (eq subform trail) (return)))))
         (loop
+         ;; circularity detection by hare and tortoise
          (frob)
          (frob)
          (setq trail (cdr trail)))))))
@@ -604,7 +600,8 @@
                       reference-constant
                       expand-compiler-macro
                       maybe-reanalyze-functional
-                      ir1-convert-common-functoid))
+                      ir1-convert-common-functoid
+                      record-macroexpand-source-path))
 
 ;;; Translate FORM into IR1. The code is inserted as the NEXT of the
 ;;; CTRAN START. RESULT is the LVAR which receives the value of the
@@ -889,6 +886,26 @@
                           (find-free-fun fun "shouldn't happen! (no-cmacro)")
                           form))))
 
+(defvar *equal-source-paths*)
+
+;;; Some macros call macroexpand-1 and then copy its results.
+(defun record-macroexpand-source-path (original-form expanded env)
+  (when (boundp '*equal-source-paths*)
+    (let ((env (if (and (boundp '*lexenv*)
+                        (or (not (lexenv-p env))
+                            (null-lexenv-p env)))
+                   *lexenv*
+                   env)))
+      (when (and (lexenv-p env)
+                 (policy env (> debug 1)))
+        (unless *equal-source-paths*
+          (setf *equal-source-paths* (make-hash-table :test #'equal)))
+        (let ((path (get-source-path original-form)))
+          (when path
+            (push path (gethash expanded *equal-source-paths*))
+            (let ((*lexenv* env))
+              (recover-source-paths original-form expanded))))))))
+
 ;;; This may produce false matches when there are multiple copies and
 ;;; they are reordered, duplicated or omitted. Still better than
 ;;; nothing.
@@ -914,7 +931,9 @@
         (let ((seen (alloc-xset)))
           (labels ((rec (form)
                      (unless (xset-member-p form seen)
-                       (let ((original (gethash form equal-table)))
+                       (let ((original (or (gethash form equal-table)
+                                           (and *equal-source-paths*
+                                                (gethash form *equal-source-paths*)))))
                          (when original
                            (let ((location
                                    (if (cdr original)
@@ -963,7 +982,8 @@
                          (t
                           (compiler-error "~@<~A~@:_ ~A~:>"
                                           (wherestring) c))))))
-      (let ((result (funcall (valid-macroexpand-hook) fun form *lexenv*)))
+      (let* (*equal-source-paths*
+             (result (funcall (valid-macroexpand-hook) fun form *lexenv*)))
         #-sb-xc-host
         (recover-source-paths form result)
         result))))
