@@ -423,47 +423,58 @@
                                      (caadr what)
                                      (second what))
                                  "-OPTIMIZER")))))
-    (if (typep what '(cons (eql vop-optimize)))
-        `(progn
-           (defun ,name (,lambda-list)
-             ,@body)
-           ,@(loop for vop-name in (ensure-list (second what))
-                   collect
-                   `(set-vop-optimizer (template-or-lose ',vop-name)
-                                       ,(if (cddr what)
-                                            `(cons #',name ',(caddr what))
-                                            `#',name))))
-        (binding* (((forms decls) (parse-body body nil))
-                   ((var-decls more-decls) (extract-var-decls decls vars))
-                   ((binds lambda-vars)
-                    (parse-deftransform lambda-list node
-                                        `(return-from ,name
-                                           ,(if (and (consp what)
-                                                     (eq (second what)
-                                                         'equality-constraint))
-                                                :give-up
-                                                nil))))
-                   (args (gensym)))
-          (declare (ignore lambda-vars))
-          `(progn
-             ;; We can't stuff the BINDS as &AUX vars into the lambda list
-             ;; because there can be a RETURN-FROM in there.
-             (defun ,name (,node ,@vars &rest ,args)
-               (declare (ignorable ,node ,@(butlast vars))
-                        (ignore ,args))
-               ,@(if var-decls (list var-decls))
-               (binding* (,@binds)
-                 (declare (ignorable ,@(loop for (bind) in binds
-                                             if (consp bind)
-                                             append bind
-                                             else
-                                             collect bind)))
-                 ,@more-decls ,@forms))
-             ,@(when (consp what)
-                 `((setf (,(let ((*package* (sb-xc:symbol-package 'fun-info)))
-                             (symbolicate "FUN-INFO-" (second what)))
-                          (fun-info-or-lose ',(first what)))
-                         #',name))))))))
+    (cond ((typep what '(cons (eql vop-optimize)))
+           `(progn
+              (defun ,name (,lambda-list)
+                ,@body)
+              ,@(loop for vop-name in (ensure-list (second what))
+                      collect
+                      `(set-vop-optimizer (template-or-lose ',vop-name)
+                                          ,(if (cddr what)
+                                               `(cons #',name ',(caddr what))
+                                               `#',name)))))
+          ((typep what '(cons t (cons (eql fold-p))))
+           `(progn
+              (defun ,name (,lambda-list)
+                ,@body)
+              (setf (,(let ((*package* (sb-xc:symbol-package 'fun-info)))
+                        (symbolicate "FUN-INFO-" (second what)))
+                     (fun-info-or-lose ',(first what)))
+                    #',name)))
+          (t
+           (binding* (((forms decls) (parse-body body nil))
+                      ((var-decls more-decls) (extract-var-decls decls vars))
+                      ((binds lambda-vars)
+                       (parse-deftransform lambda-list node
+                                           `(return-from ,name
+                                              ,(if (and (consp what)
+                                                        (eq (second what)
+                                                            'equality-constraint))
+                                                   :give-up
+                                                   nil))))
+                      (args (gensym)))
+             (declare (ignore lambda-vars))
+             `(progn
+                ;; We can't stuff the BINDS as &AUX vars into the lambda list
+                ;; because there can be a RETURN-FROM in there.
+                (defun ,name (,node ,@vars &rest ,args)
+                  (declare (ignorable ,node ,@(remove-if (lambda (v)
+                                                           (member v lambda-list-keywords))
+                                                         (butlast vars)))
+                           (ignore ,args))
+                  ,@(if var-decls (list var-decls))
+                  (binding* (,@binds)
+                    (declare (ignorable ,@(loop for (bind) in binds
+                                                if (consp bind)
+                                                append bind
+                                                else
+                                                collect bind)))
+                    ,@more-decls ,@forms))
+                ,@(when (consp what)
+                    `((setf (,(let ((*package* (sb-xc:symbol-package 'fun-info)))
+                                (symbolicate "FUN-INFO-" (second what)))
+                             (fun-info-or-lose ',(first what)))
+                            #',name)))))))))
 
 (defmacro defoptimizers (kind names (lambda-list
                                      &optional (node (gensym))
@@ -640,7 +651,10 @@
   (once-only ((node node))
     `(let ((*current-component* (node-component ,node))
            (*lexenv* (node-lexenv ,node))
-           (*current-path* (node-source-path ,node)))
+           (*current-path* (node-source-path ,node))
+           (*inline-expansions* (if (basic-combination-p ,node)
+                                    (basic-combination-inline-expansions ,node)
+                                    *inline-expansions*)))
        (aver-live-component *current-component*)
        ,@forms)))
 
@@ -1012,15 +1026,39 @@ specify bindings for printer control variables.")
 (defmacro make-defs (vars &body body)
   (labels ((subst-if-with (test tree)
              (labels ((s (subtree)
-                        (let ((test (funcall test subtree)))
-                          (cond (test)
+                        (multiple-value-bind (new replace)
+                            (funcall test subtree)
+                          (cond (replace new)
+                                ((comma-p subtree)
+                                 (let ((new (s (comma-expr subtree))))
+                                   (if (eq subtree new)
+                                       subtree
+                                       (unquote new (comma-kind subtree)))))
                                 ((atom subtree) subtree)
                                 (t (let ((car (s (car subtree)))
                                          (cdr (s (cdr subtree))))
                                      (if (and (eq car (car subtree))
                                               (eq cdr (cdr subtree)))
                                          subtree
-                                         (cons car cdr))))))))
+                                         (cond ((and (typep car '(cons symbol))
+                                                     (string= (car car) "$WHEN"))
+                                                (if (eval (second car))
+                                                    (append (cddr car) cdr)
+                                                    cdr))
+                                               ((and (typep car '(cons symbol))
+                                                     (string= (car car) "$UNLESS"))
+                                                (if (eval (second car))
+                                                    cdr
+                                                    (append (cddr car) cdr)))
+                                               ((and (typep car '(cons symbol))
+                                                     (string= (car car) "$IF"))
+                                                (cons
+                                                 (if (eval (second car))
+                                                     (third car)
+                                                     (fourth car))
+                                                 cdr))
+                                               (t
+                                                (cons car cdr))))))))))
                (s tree)))
            (test (pattern with)
              (lambda (x)
@@ -1029,10 +1067,11 @@ specify bindings for printer control variables.")
                         (start (search pattern str)))
                    (when start
                      (if (equal str pattern)
-                         with
-                         (symbolicate (subseq str 0 start)
-                                      with
-                                      (subseq str (+ start (length pattern))))))))))
+                         (values with t)
+                         (values (symbolicate (subseq str 0 start)
+                                              with
+                                              (subseq str (+ start (length pattern))))
+                                 t)))))))
            (gen (vars body)
              (if vars
                  (loop for with in (cdar vars)
