@@ -124,7 +124,10 @@
              ((nil) 'unsigned-byte)
              ((t) 'signed-byte))
           ,width)
-         (foldable flushable movable always-translatable)
+         (foldable flushable movable always-translatable
+                   ,@(and (ir1-attributep (fun-info-attributes (fun-info-or-lose prototype))
+                                          commutative)
+                          '(commutative)))
        :derive-type (make-modular-fun-type-deriver ',prototype ,width ',signedp))))
 
 (defun %define-good-modular-fun (name kind signedp)
@@ -141,19 +144,24 @@
                                                  (node (gensym "NODE")))
      &body body)
   (%check-modular-fun-macro-arguments name kind lambda-list)
-  (with-unique-names (args result-width-name)
-    `(setf (gethash ',name (modular-class-funs (find-modular-class ',kind ',signedp)))
-           (lambda (,node ,width ,(or result-width
-                                   result-width-name))
-             (declare (type basic-combination ,node)
-                      (type (integer 0) ,width)
-                      ,@(unless result-width
-                          `((ignore ,result-width-name))))
-             (let ((,args (basic-combination-args ,node)))
-               (when (= (length ,args) ,(length lambda-list))
-                 (destructuring-bind ,lambda-list ,args
-                   (declare (type lvar ,@lambda-list))
-                   ,@body)))))))
+  (multiple-value-bind (forms decls) (parse-body body nil)
+    (multiple-value-bind (a-declarations b-declarations)
+        (extract-var-decls decls (and result-width (list result-width)))
+      (with-unique-names (args result-width-name)
+        `(setf (gethash ',name (modular-class-funs (find-modular-class ',kind ',signedp)))
+               (lambda (,node ,width ,(or result-width
+                                       result-width-name))
+                 (declare (type basic-combination ,node)
+                          (type (integer 0) ,width)
+                          ,@(unless result-width
+                              `((ignore ,result-width-name))))
+                 ,a-declarations
+                 (let ((,args (basic-combination-args ,node)))
+                   (when (= (length ,args) ,(length lambda-list))
+                     (destructuring-bind ,lambda-list ,args
+                       (declare (type lvar ,@lambda-list))
+                       ,@b-declarations
+                       ,@forms)))))))))
 
 ;;; (ldb (byte s 0) (foo                 x  y ...)) =
 ;;; (ldb (byte s 0) (foo (ldb (byte s 0) x) y ...))
@@ -193,17 +201,16 @@
                           (args (combination-args dest)))
                  (case name
                    (logand
-                    (when (= 2 (length args))
-                      (let ((other (if (eql (first args) lvar)
-                                       (second args)
-                                       (first args))))
+                    (when (and (= (length args) 2)
+                               (eq (first args) lvar))
+                      (let ((other (second args)))
                         (when (and (constant-lvar-p other)
-                                   (ctypep (lvar-value other) type)
-                                   (not signedp))
+                                   (typep (lvar-value other) 'unsigned-byte)
+                                   (ctypep (lvar-value other) type))
                           (return-from insert-lvar-cut)))))
                    (mask-signed-field
                     (when (and signedp
-                               (eql lvar (second args))
+                               (eq lvar (second args))
                                (constant-lvar-p (first args))
                                (<= (lvar-value (first args)) width))
                       (return-from insert-lvar-cut)))))
@@ -446,23 +453,136 @@
               nil                ; After fixing above, replace with T.
               )))))))
 
+(defun logior-cut-width (node)
+  (multiple-value-bind (low high)
+      (integer-type-numeric-bounds (single-value-result-type node))
+    (when (and (numberp low)
+               (numberp high)
+               (<= high 0))
+      (let ((width (integer-length low)))
+        (best-modular-version (1+ width) t)))))
+
 (defoptimizer (logior optimizer) ((x y) node)
-  (let ((result-type (single-value-type (node-derived-type node))))
-    (multiple-value-bind (low high)
-        (integer-type-numeric-bounds result-type)
-      (when (and (numberp low)
-                 (numberp high)
-                 (<= high 0))
-        (let ((width (integer-length low)))
-          (multiple-value-bind (w kind)
-              (best-modular-version (1+ width) t)
-            (when w
-              ;; FIXME: see comment in LOGAND optimizer
-              (let ((xact (cut-to-width x kind w t))
-                    (yact (cut-to-width y kind w t)))
-                (declare (ignore xact yact))
-                nil) ; After fixing above, replace with T
-              )))))))
+  (multiple-value-bind (w kind)
+      (logior-cut-width node)
+    (when w
+      ;; FIXME: see comment in LOGAND optimizer
+      (let ((xact (cut-to-width x kind w t))
+            (yact (cut-to-width y kind w t)))
+        (declare (ignore xact yact))
+        nil) ; After fixing above, replace with T
+      )))
+
+
+(defun unsigned-mask-width (type)
+  (let* ((int (type-approximate-interval type))
+         (high (interval-high int)))
+    (when high
+      (integer-length high))))
+
+(deftransform logand ((x y) (t (constant-arg integer)) word
+                      :node node :important nil)
+  ;; Reduce constant width
+  (let* ((mask (lvar-value y))
+         (cut (ldb (byte (unsigned-mask-width (single-value-result-type node t)) 0)
+                   mask)))
+    (if (= cut mask)
+        (give-up-ir1-transform)
+        `(logand x ,cut))))
+
+;;; Remove the second logand or reduce its constant in
+;;; (logand m (logand n #xFFFF))
+(deftransform logand ((a b) (t t) * :important nil :node node)
+  (or (combination-match (:node node)
+          (logand (:type unsigned-byte a) (logand x (:constant b)))
+        (block nil
+          (let* ((width (or (unsigned-mask-width (lvar-type a))
+                            (return)))
+                 (full-mask (if (constant-lvar-p a)
+                                (lvar-value a)
+                                (ldb (byte width 0) -1)))
+                 (cut (logand b
+                              full-mask)))
+            (cond ((and
+                    ;; unsigned cut-to-width always recuts to the minimum width
+                    (vop-existsp :translate sb-vm::*-modfx)
+                    ;; cut-to-width will insert these again
+                    (/= cut most-positive-word
+                        (ash most-positive-word -1))
+                    (= cut full-mask))
+                   (extract-lvar-n x 1 node)
+                   t)
+                  ((= cut b)
+                   nil)
+                  (t
+                   (erase-node-type combination *wild-type* nil node)
+                   (transform-call combination
+                                   `(lambda (x y)
+                                      (declare (ignore y))
+                                      (logand x ,cut))
+                                   'logand)
+                   t)))))
+      ;; Reduce the constant in logior
+      (combination-match (:node node)
+          (logand (:type unsigned-byte a) (logior * (:constant b)))
+        (block nil
+          (let* ((width (or (unsigned-mask-width (lvar-type a))
+                            (return)))
+                 (full-mask (ldb (byte width 0) -1))
+                 (mask (if (constant-lvar-p a)
+                           (lvar-value a)
+                           full-mask))
+                 (cut (logand b mask)))
+            (cond ((= cut full-mask)
+                   ;; (logand #xFF (logior n #xFF)) => #xFF
+                   (erase-node-type combination *wild-type* nil node)
+                   (transform-call combination
+                                   `(lambda (x y)
+                                      (declare (ignore x y))
+                                      ,cut)
+                                   'logand)
+                   t)
+                  ((or (>= (integer-length cut)
+                           (integer-length b)))
+                   nil)
+                  (t
+                   (erase-node-type combination *wild-type* nil node)
+                   (transform-call combination
+                                   `(lambda (x y)
+                                      (declare (ignore y))
+                                      (logior x ,cut))
+                                   'logand)
+                   t)))))
+      (combination-match (:node node)
+          (logand (:type unsigned-byte a) (logxor * (:constant b)))
+        (block nil
+          (let* ((width (or (unsigned-mask-width (lvar-type a))
+                            (return)))
+                 (full-mask (ldb (byte width 0) -1))
+                 (mask (if (constant-lvar-p a)
+                           (lvar-value a)
+                           full-mask))
+                 (cut (logand b
+                              mask)))
+            (cond ((= cut b)
+                   nil)
+                  (t
+                   (erase-node-type combination *wild-type* nil node)
+                   (transform-call combination
+                                   `(lambda (x y)
+                                      (declare (ignore y))
+                                      (logxor x ,cut))
+                                   'logand)
+                   t)))))
+      ;; Remove mask-signed-field
+      (combination-match (:node node)
+          (logand (:type unsigned-byte a) (mask-signed-field (:constant sign) b))
+        (block nil
+          (let ((width (or (unsigned-mask-width (lvar-type a))
+                           (return))))
+            (when (> sign width)
+              (extract-lvar-n b 1 node))))))
+  (give-up-ir1-transform))
 
 ;;; Combine (ash (ash x 1) 1) into (ash x 2)
 (deftransform ash ((value amount))
@@ -513,6 +633,141 @@
               `(lambda (value amount1 amount2)
                  (logand (ash value (+ amount1 amount2))
                          ,(1- (ash 1 (+ width (lvar-value amount))))))))))))
+
+(deftransform ash-into-word-mod ((x count) (t (constant-arg (integer #.(- sb-vm:n-word-bits) 0))))
+  `(typecase x
+     (fixnum
+      (logand most-positive-word (ash (truly-the fixnum x) count)))
+     (sb-vm:signed-word
+      (logand most-positive-word (ash (truly-the sb-vm:signed-word x) count)))
+     (t
+      (ash-right-two-words (sb-bignum:%bignum-ref (truly-the bignum x) 1)
+                           (sb-bignum:%bignum-ref (truly-the bignum x) 0)
+                           (- count)))))
+
+(deftransform ash-into-word-mod ((x count) * * :important nil :priority :last)
+  (let* ((minusp (csubtypep (lvar-type count) (specifier-type '(integer * 0))))
+         (right
+           `(typecase x
+              (fixnum
+               (logand most-positive-word (ash (truly-the fixnum x) ,(if minusp
+                                                                         `(- count)
+                                                                         `count))))
+              (t
+               (multiple-value-bind (words bits) (truncate ,(if minusp
+                                                                `count
+                                                                `(- count))
+                                                           sb-vm:n-word-bits)
+                 (let ((length (%bignum-length (truly-the bignum x))))
+                   (flet ((extend-ref (index)
+                            (if (< index length)
+                                (sb-bignum:%bignum-ref (truly-the bignum x) (truly-the bignum-index index))
+                                (sb-bignum::%sign-digit (truly-the bignum x) length))))
+                     (declare (inline extend-ref))
+                     (ash-right-two-words (extend-ref (1+ words))
+                                          (extend-ref words)
+                                          (truly-the (mod ,sb-vm:n-word-bits) bits)))))))))
+    (if minusp
+        `(let ((count (- count)))
+           ,right)
+        `(if (minusp count)
+             ,right
+             (logand most-positive-word (ash x (truly-the unsigned-byte count)))))))
+
+(defun ash-into-word-mod (x count)
+  (logand (ash x count) most-positive-word))
+
+(defun ash-into-word-modfx (x count)
+  (mask-signed-field sb-vm:n-fixnum-bits (ash x count)))
+
+(deftransform ash-into-word-mod ((x count) (t (constant-arg (integer * (#.(- sb-vm:n-word-bits))))))
+  (let ((count (- (lvar-value count))))
+    (multiple-value-bind (words bits) (truncate count sb-vm:n-word-bits)
+      `(typecase x
+         (sb-vm:signed-word
+          (logand most-positive-word (ash (truly-the sb-vm:signed-word x) count)))
+         (t
+          (let ((length (%bignum-length (truly-the bignum x))))
+            (flet ((extend-ref (index)
+                     (if (< index length)
+                         (sb-bignum:%bignum-ref (truly-the bignum x) index)
+                         (sb-bignum::%sign-digit (truly-the bignum x) length))))
+              (declare (inline extend-ref))
+              ,(if (zerop bits)
+                   `(extend-ref ,words)
+                   `(ash-right-two-words (extend-ref ,(1+ words))
+                                         (extend-ref ,words)
+                                         ,bits)))))))))
+
+(deftransform ash-into-word-modfx ((x count) (t (constant-arg (integer #.(- sb-vm:n-word-bits) 0))))
+  `(typecase x
+     (fixnum
+      (mask-signed-field sb-vm:n-fixnum-bits
+                         (ash (truly-the fixnum x) count)))
+     (sb-vm:signed-word
+      (mask-signed-field sb-vm:n-fixnum-bits
+                         (ash (truly-the sb-vm:signed-word x) count)))
+     (t
+      (mask-signed-field sb-vm:n-fixnum-bits
+                         (ash-right-two-words (sb-bignum:%bignum-ref (truly-the bignum x) 1)
+                                              (sb-bignum:%bignum-ref (truly-the bignum x) 0)
+                                              (- count))))))
+
+(deftransform ash-into-word-modfx ((x count) * * :important nil :priority :last)
+  (let* ((minusp (csubtypep (lvar-type count) (specifier-type '(integer * 0))))
+         (right
+           `(typecase x
+              (fixnum
+               (mask-signed-field sb-vm:n-fixnum-bits
+                                  (ash (truly-the fixnum x) ,(if minusp
+                                                                 `(- count)
+                                                                 `count))))
+              (t
+               (multiple-value-bind (words bits) (truncate ,(if minusp
+                                                                `count
+                                                                `(- count))
+                                                           sb-vm:n-word-bits)
+                 (let ((length (%bignum-length (truly-the bignum x))))
+                   (flet ((extend-ref (index)
+                            (if (< index length)
+                                (sb-bignum:%bignum-ref (truly-the bignum x) (truly-the bignum-index index))
+                                (sb-bignum::%sign-digit (truly-the bignum x) length))))
+                     (declare (inline extend-ref))
+                     (mask-signed-field sb-vm:n-fixnum-bits
+                                        (ash-right-two-words (extend-ref (1+ words))
+                                                             (extend-ref words)
+                                                             bits)))))))))
+    (if minusp
+        `(let ((count (- count)))
+           ,right)
+        `(if (minusp count)
+             ,right
+             (mask-signed-field sb-vm:n-fixnum-bits (ash x (truly-the unsigned-byte count)))))))
+
+(deftransform ash-into-word-modfx ((x count) (t (constant-arg (integer * (#.(- sb-vm:n-word-bits))))))
+  (let ((count (- (lvar-value count))))
+    (multiple-value-bind (words bits) (truncate count sb-vm:n-word-bits)
+      `(typecase x
+         (sb-vm:signed-word
+          (mask-signed-field sb-vm:n-fixnum-bits
+                             (ash (truly-the sb-vm:signed-word x) count)))
+         (t
+          (let ((length (%bignum-length (truly-the bignum x))))
+            (flet ((extend-ref (index)
+                     (if (< index length)
+                         (sb-bignum:%bignum-ref (truly-the bignum x) index)
+                         (sb-bignum::%sign-digit x length))))
+              (declare (inline extend-ref))
+              (mask-signed-field sb-vm:n-fixnum-bits
+                                 ,(if (zerop bits)
+                                      `(extend-ref ,words)
+                                      `(ash-right-two-words (extend-ref ,(1+ words))
+                                                            (extend-ref ,words)
+                                                            ,bits))))))))))
+
+(deftransform ash-right-two-words ((w2 w1 count) (t t (eql #.sb-vm:n-word-bits)) * :important nil)
+  'w2)
+
 (macrolet
     ((def (left-name name kind width signedp)
        (declare (ignorable name))
@@ -525,6 +780,7 @@
               :derive-type (make-modular-fun-type-deriver 'ash ',width ',signedp))
             (define-modular-fun-optimizer ash ((integer count) ,kind ,signedp :width width
                                                :result-width result-width)
+              (declare (ignorable result-width))
               (let ((integer-type (lvar-type integer))
                     (count-type (lvar-type count)))
                 (declare (ignorable integer-type))
@@ -540,14 +796,19 @@
                               ;; Unknown sign
                               (not (csubtypep count-type (specifier-type '(integer * 0))))
                               (not (csubtypep count-type (specifier-type '(integer 0 *))))
-                              (or (csubtypep integer-type (specifier-type `(unsigned-byte ,sb-vm:n-word-bits)))
-                                  (csubtypep integer-type (specifier-type `(signed-byte ,sb-vm:n-word-bits)))))
+                              (word-sized-type-p integer-type))
                          ',name)
-                        ((and (not (csubtypep integer-type (specifier-type 'word)))
-                              (not (csubtypep integer-type (specifier-type 'sb-vm:signed-word)))
-                              (csubtypep count-type (specifier-type `(integer ,(- result-width width) ,most-positive-fixnum))))
-                         (cut-to-width integer ,kind width ,signedp)
-                         t)))))
+                        #+(or arm64 x86-64)
+                        ((not (word-sized-type-p integer-type))
+                         (cond ((csubtypep count-type (specifier-type `(integer ,(- result-width width) ,most-positive-fixnum)))
+                                ;; Uses the bits from the first word when shifting right
+                                (cut-to-width integer ,kind width ,signedp)
+                                ',name)
+                               ((and (csubtypep count-type (specifier-type 'fixnum))
+                                     (not (csubtypep count-type (specifier-type 'unsigned-byte))))
+                                ',(if signedp
+                                      'ash-into-word-modfx
+                                      'ash-into-word-mod))))))))
             (setf (gethash ',left-name (modular-class-versions (find-modular-class ',kind ',signedp)))
                   `(ash ,',width))
             (deftransform ,left-name ((integer count) (t (constant-arg (eql 0))))
